from os import listdir, makedirs, replace, rmdir, rename
from os.path import splitext, basename, isfile, join, isdir
            

class DirectoryDecoratory():
    # Extensions
    audio_ext = [".mp3", ".ogg", ".wav", ".m4a"]
    video_ext = [".mp4", ".fs", ".mkv", ".avi"]
    image_ext = [".jpg", ".jpeg", ".png", ".ico", ".gif", ".tiff"]
    document_ext = [".pdf", ".txt", ".docx", ".doc", ".odt"]
    spreadsheet_ext = [".xlsx", ".xls", ".odp"]
    presentation_ext = [".pptx", ".ppt", ".odp"]
    archive_ext = [".zip", ".rar", ".tar", ".7z", ".iso"]
    application_ext = [".exe", ".msi", ".apk", ".xapk"]
    script_ext = [".py", ".pyc", ".pyw", ".c", ".cs", ".cpp"]
    database_ext = [".sql", ".db", ".dbf", ".odb", ".accdb", ".odf"]
    binary_ext = [".bin", ".pkl"]
    web_ext = [".html", ".htm", ".css", ".js", ".json"]

    def __init__(self) -> None:
        self.program_running = True

        self.folder_path = ""
        self.files = []

        self.categories = {
            "Audios": [],
            "Videos": [],
            "Images": [],
            "Documents": [],
            "Spreadsheets": [],
            "Presentations": [],
            "Archives": [],
            "Applications": [],
            "Scripts": [],
            "Databases": [],
            "Binaries": [],
            "Web": [],
            "Others": []
        }

        self.new_and_old_file_names = {}  # new_file_name: old_file_name
        self.exceptions = []
        self.extensions_for_numbering = {}

    def to_exit(self, input_value: str) -> bool:
        if input_value.lower() == "q":
            print("\nThanks for using this tool!")
            self.program_running = False
            return True

        return False
    
    # This method will ask the user to input the folder path of which he/she wants to decorate
    def get_folder_path(self) -> None:
        while True:
            self.folder_path = input("\nEnter the path of the folder you want to decorate: ").strip()
                            
            if self.folder_path == "":
                continue
            elif self.to_exit(self.folder_path):
                break
            elif isdir(self.folder_path):
                files_and_folders = listdir(self.folder_path)  # list of all files and folders

                # __file__ is the path of this python program file
                # If this python program exists in the folder provided by the user, remove it from the above list
                # to prevent it interfering with itself
                if __file__.lower() == join(self.folder_path, basename(__file__)).lower():
                    files_and_folders.remove(basename(__file__))  # basename(__file__) returns python file's name

                self.files = [file_name for file_name in files_and_folders if isfile(join(self.folder_path, file_name))]
                if len(self.files) == 0:
                    print(f"This directory doesn't contain any files!")
                else:
                    break
            else:
                print("No such directory! Please enter a valid path.")
    
    # This method will add the files of the folder in respective lists based on their file extensions
    def categorize_files(self) -> None:
        for file in self.files:
            extension = splitext(file)[1].lower()
            if extension in self.audio_ext:
                self.categories["Audios"].append(file)
            elif extension in self.video_ext:
                self.categories["Videos"].append(file)
            elif extension in self.image_ext:
                self.categories["Images"].append(file)
            elif extension in self.document_ext:
                self.categories["Documents"].append(file)
            elif extension in self.spreadsheet_ext:
                self.categories["Spreadsheets"].append(file)
            elif extension in self.presentation_ext:
                self.categories["Presentations"].append(file)
            elif extension in self.archive_ext:
                self.categories["Archives"].append(file)
            elif extension in self.application_ext:
                self.categories["Applications"].append(file)
            elif extension in self.script_ext:
                self.categories["Scripts"].append(file)
            elif extension in self.database_ext:
                self.categories["Databases"].append(file)
            elif extension in self.binary_ext:
                self.categories["Binaries"].append(file)
            elif extension in self.web_ext:
                self.categories["Web"].append(file)
            else:
                self.categories["Others"].append(file)
    
    # This method will create the folders and move the respective files in them
    def clear_clutter(self) -> None:
        for folder, files in self.categories.items():
            if len(files) > 0:
                makedirs(join(self.folder_path, folder), exist_ok=True)  # creates the folder
                for file in files:
                    replace(join(self.folder_path, file), join(self.folder_path, folder, file))  # move the files
    
    # This method will undo the clearing of clutter by moving files to their original path
    def undo_clutter(self) -> None:
        for folder in self.categories.keys():
            if len(self.categories[folder]) > 0:
                for file in listdir(join(self.folder_path, folder)):
                    # moves the file to its original path
                    replace(join(self.folder_path, folder, file), join(self.folder_path, file))

                try:
                    rmdir(join(self.folder_path, folder))  # delete the category folder
                except PermissionError:
                    continue

    # This method will ask the user for extensions which he/she wants to ignore while renaming files
    def get_exceptions(self) -> None:
        while True:
            exceptions = input("Enter comma-separated extensions of files you want to ignore (leave blank if you don't want to ignore any): ").strip()

            if self.to_exit(exceptions) or exceptions == "":
                break
            else:
                for extension in exceptions.split():
                    if "." not in extension:
                        extension = "." + extension

                    self.exceptions.append(extension.replace(" ", "").lower())
                break

    # This method will ask the user for extensions on which he/she wants to apply numbering while renaming 
    def get_extensions_for_numbering(self) -> None:
        while True:
            extensions = input("Enter comma-separated extensions of files you want to apply numbering on (leave blank if you don't want numbering): ").strip()

            if self.to_exit(extensions) or extensions == "":
                break
            else:
                for extension in extensions.split(','):
                    if "." not in extension:
                        extension = "." + extension

                    self.extensions_for_numbering[extension.replace(" ", "").lower()] = 1
                break
    
    # This method checks if the extension of a file is marked to apply numbering
    def apply_numbering(self, extension: str) -> bool:
        if extension in self.extensions_for_numbering:
            print(f"{extension} is marked for numbering.")
            return True
        
        return False
    
    # This method checks if the extension is meant to be ignored
    def check_exception(self, file_name: str, extension: str) -> bool:
        if extension not in self.exceptions:
            return True
        
        print(f"{extension} is marked as exception: Skipping '{file_name}'")
        return False

    # This method will rename the files and also handle the numbering and ignoring of files
    def rename_files(self) -> None:
        print("\n----------------------------------")
        print("RENAMING FILES...\n")
        for full_file_name in self.files:
            file_name = splitext(full_file_name)[0].title()
            extension = splitext(full_file_name)[1].lower()
            new_file_name = file_name + extension

            if self.apply_numbering(extension):
                number = str(self.extensions_for_numbering[extension])
                rename(join(self.folder_path, full_file_name), join(self.folder_path, number + extension))
                print(f"RENAMED:   {full_file_name} ---> {number + extension}")
                self.new_and_old_file_names[number + extension] = full_file_name
                self.extensions_for_numbering[extension] += 1
            elif self.check_exception(file_name, extension):
                rename(join(self.folder_path, full_file_name), join(self.folder_path, new_file_name))
                print(f"RENAMED:   {full_file_name} ---> {new_file_name}")
                self.new_and_old_file_names[new_file_name] = full_file_name
            
        print("\nRENAMED ALL FILES.")
        print("----------------------------------\n")
    
    # This method will rename the files back to their original names
    def undo_prettify(self) -> None:
        for new_name, old_name in self.new_and_old_file_names.items():
            rename(join(self.folder_path, new_name), join(self.folder_path, old_name))

    # This method will ask the user what he/she wants to perform on the provided folder
    def what_to_do(self) -> str:
        print("\nPlease choose what you want to perform on this folder:")
        print("1. Clear The Clutter")
        print("\tOrganize the files in folders with respect to their extension types")
        print("2. Prettify The Folder")
        print("\tRename files to make the folder look clean\n")

        while True:
            to_do = input("Choose option (1/2): ").strip()
            if self.to_exit(to_do):
                break
            else:
                if to_do == "1":
                    return "clear"
                elif to_do == "2":
                    return "prettify"
                else:
                    print("Invalid input! Please try again...\n")

    # This method starts the main program
    def start(self) -> None:
        print("----- Directory Decoratory -----")
        print("Anywhere in the program, enter q to exit")

        while self.program_running:
            self.get_folder_path()

            if self.program_running: 
                to_do = self.what_to_do()

            if self.program_running:
                if to_do == "clear":
                    self.categorize_files()
                    self.clear_clutter()
                    print("Success! The clutter of your files in now cleared. Check them now.\n")
                    while True:
                        want_undo = input("If you are not satisfied with the results, you can undo now. Do you want to undo? (y/n): ").strip().lower()
                        if want_undo == "y":
                            self.undo_clutter()
                            print("Success! Your files are restored to their original path.")
                            break
                        elif want_undo == "n":
                            break
                        else:
                            print("Invalid input! Please try again...\n")
                elif to_do == "prettify":
                    if self.program_running: 
                        self.get_exceptions()
                    if self.program_running: 
                        self.get_extensions_for_numbering()
                    if self.program_running: 
                        self.rename_files()
                        while True:
                            want_undo = input("If you are not satisfied with the results, you can undo now. Do you want to undo? (y/n): ").strip().lower()
                            if want_undo == "y":
                                self.undo_prettify()
                                print("Success! Your files are renamed to their original names.")
                                break
                            elif want_undo == "n":
                                break
                            else:
                                print("Invalid input! Please try again...\n")


if __name__ == "__main__":
    decorator = DirectoryDecoratory()
    decorator.start()
